'''
Code-behind the scenes for the cli application.
'''
import json
import time

import boto3
from botocore.exceptions import ClientError
import click
import requests
from bs4 import BeautifulSoup

from clokta.common import Common
from clokta.factor_chooser import FactorChooser
from clokta.role_chooser import RoleChooser
from clokta.profile_manager import ProfileManager


class RoleAssumer(object):
    ''' Core implementation of clokta '''

    def __init__(self, profile, verbose=False):
        ''' Constructor '''
        self.profile = profile
        self.verbose = verbose

    def assume_role(self):
        ''' entry point for the cli tool '''
        profile_mgr = ProfileManager(profile_name=self.profile, verbose=self.verbose)
        configuration = profile_mgr.initialize_configuration()
        profile_mgr.update_configuration(profile_configuration=configuration)

        session_token = self.__okta_session_token(
            configuration=configuration
        )
        if self.verbose:
            Common.dump_verbose(message='Okta session token: {}'.format(session_token))

        saml_assertion = self.__saml_assertion_aws(
            session_token=session_token,
            configuration=configuration
        )

        idp_and_role_chooser = RoleChooser(
            saml_assertion=saml_assertion,
            role_preference=configuration.get('okta_aws_role_to_assume'),
            verbose=self.verbose
        )
        idp_role_tuple = idp_and_role_chooser.choose_idp_role_tuple()

        client = boto3.client('sts')
        # Try for a 12 hour session.  If it fails, try for a 1 hour session
        try:
            assumed_role_credentials = client.assume_role_with_saml(
                RoleArn=idp_role_tuple[2],
                PrincipalArn=idp_role_tuple[1],
                SAMLAssertion=saml_assertion,
                DurationSeconds=43200
            )
        except ClientError as e:
            if e.response['Error']['Code'] == 'ValidationError':
                assumed_role_credentials = client.assume_role_with_saml(
                    RoleArn=idp_role_tuple[2],
                    PrincipalArn=idp_role_tuple[1],
                    SAMLAssertion=saml_assertion,
                    DurationSeconds=3600
                )
                Common.echo(message='YOUR SESSION WILL ONLY LAST ONE HOUR')
            else:
                raise

        profile_mgr.apply_credentials(credentials=assumed_role_credentials, echo_message=True)
        bash_file = profile_mgr.write_sourceable_file(credentials=assumed_role_credentials)
        docker_file = profile_mgr.write_dockerenv_file(credentials=assumed_role_credentials)
        Common.echo(
            message='AWS keys generated. To use, run "export AWS_PROFILE={prof}"\nor use generated files {file1} with docker compose or {file2} with shell scripts'.format(
                prof=self.profile, file1=docker_file, file2=bash_file
            )
        )

    def __okta_session_token(self, configuration):
        ''' Authenticate with Okta; receive a session token '''
        okta_response = None

        try:
            okta_response = self.__okta_auth_response(configuration=configuration)
        except requests.exceptions.HTTPError as http_err:
            if self.verbose:
                msg = 'Okta returned this credentials/password related error: {}\nThis could be a mistyped password or a misconfigured username or URL.'.format(http_err)
            else:
                msg = "Failure.  Wrong password or misconfigured session."
            Common.dump_err(message=msg, exit_code=1, verbose=self.verbose)
        except Exception as err:
            msg = 'Unexpected error: {}'.format(err)
            Common.dump_err(message=msg, exit_code=2, verbose=self.verbose)

        # handle case where MFA is required but no factors have been enabled
        if okta_response['status'] == 'MFA_ENROLL':
            msg = 'Please enroll in multi-factor authentication before using this tool'
            Common.dump_err(message=msg, exit_code=3, verbose=self.verbose)

        otp_value = None
        if configuration.get('okta_onetimepassword_secret'):
            try:
                import onetimepass as otp
            except ImportError:
                msg = 'okta_onetimepassword_secret provided in config but "onetimepass" is not installed. run: pip install onetimepass'
                Common.dump_err(message=msg, exit_code=3, verbose=self.verbose)
            otp_value = otp.get_totp(configuration['okta_onetimepassword_secret'])

        if okta_response['status'] == 'MFA_REQUIRED':
            factors = okta_response['_embedded']['factors']
            if factors:
                return self.__okta_session_token_mfa(
                    auth_response=okta_response,
                    factors=factors,
                    factor_preference=configuration['multifactor_preference'],
                    otp_value = otp_value
                )
            else:
                msg = 'No MFA factors have been set up for this account'
                Common.dump_err(message=msg, exit_code=3, verbose=self.verbose)

        return okta_response['sessionToken']

    def __okta_session_token_mfa(self, auth_response, factors, factor_preference, otp_value=None):
        ''' Determine which factor to use and apply it to get a session token '''
        factor = self.__choose_factor(
            factors=factors,
            factor_preference=factor_preference,
        )
        state_token = auth_response['stateToken']

        if factor['factorType'] == 'push':
            return self.__send_push(
                factor=factor,
                state_token=state_token
            )

        if factor['factorType'] == 'sms':
            self.__okta_mfa_verification(
                factor_dict=factor,
                state_token=state_token,
                otp_value=None
            )

        if not otp_value:
            otp_value = click.prompt('Enter your multifactor authentication token', type=str)
        try:
            mfa_response = self.__okta_mfa_verification(
                factor_dict=factor,
                state_token=state_token,
                otp_value=otp_value
            )
            session_token = mfa_response['sessionToken']
        except requests.exceptions.HTTPError as http_err:
            msg = 'Okta returned this MFA related error: {}'.format(http_err)
            Common.dump_err(message=msg, exit_code=1, verbose=self.verbose)
        except Exception as err:
            msg = 'Unexpected error: {}'.format(err)
            Common.dump_err(message=msg, exit_code=2, verbose=self.verbose)

        return session_token

    def __send_push(self, factor, state_token):
        ''' Send push re: Okta Verify '''
        url = factor['_links']['verify']['href']
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache'
        }
        payload = {
            'stateToken': state_token
        }

        response_data = None
        response = requests.post(url, data=json.dumps(payload), headers=headers)
        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            response_data = response.json()
        else:
            response.raise_for_status()

        Common.echo(message='Push notification sent; waiting for your response', new_line=False)

        status = response_data['status']
        if status == 'MFA_CHALLENGE':
            if 'factorResult' in response_data and response_data['factorResult'] == 'WAITING':
                return self.__check_push_result(
                    state_token=state_token,
                    push_response=response_data
                )

    def __check_push_result(self, state_token, push_response):
        ''' Wait for push response acknowledgement '''
        url = push_response['_links']['next']['href']
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache'
        }
        payload = {
            'stateToken': state_token
        }

        wait_for = 60
        timeout = time.time() + wait_for
        response_data = None
        while True:
            Common.echo(message='.', new_line=False)
            response = requests.post(url, data=json.dumps(payload), headers=headers)
            if response.status_code == requests.codes.ok:  # pylint: disable=E1101
                response_data = response.json()
            else:
                response.raise_for_status()

            if 'sessionToken' in response_data or time.time() > timeout:
                Common.echo(message='Session confirmed')
                break
            time.sleep(3)

        if response_data:
            return response_data['sessionToken']
        else:
            msg = 'Timeout expired ({} seconds)'.format(wait_for)
            Common.dump_err(message=msg, exit_code=3)

    def __choose_factor(self, factors, factor_preference=None):
        ''' Automatically choose, or allow user to choose, the MFA option '''

        fact_chooser = FactorChooser(
            factors=factors,
            factor_preference=factor_preference,
            verbose=self.verbose
        )

        # Is there only one legitimate MFA option available?
        if len(factors) == 1:
            factor = fact_chooser.verify_only_factor(factor=factors[0])
            if factor:
                return factor

        # Has the user pre-selected a legitimate factor?
        if factor_preference:
            factor = fact_chooser.verify_preferred_factor()
            if factor:
                return factor

        return fact_chooser.choose_supported_factor()

    def __okta_mfa_verification(self, factor_dict, state_token, otp_value=None):
        ''' Sends the MFA token entered and retuns the response '''
        url = factor_dict['_links']['verify']['href']
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache'
        }
        payload = {
            'stateToken': state_token
        }
        if otp_value:
            payload['answer'] = otp_value

        response = requests.post(url, data=json.dumps(payload), headers=headers)
        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            return response.json()
        else:
            response.raise_for_status()

    def __okta_auth_response(self, configuration):
        ''' Returns an HTTP response for credentials-based authentication with Okta '''
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache',
            'Authorization': 'API_TOKEN'
        }
        payload = {
            'username': configuration['okta_username'],
            'password': configuration['okta_password']
        }
        url = 'https://' + configuration['okta_org'] + '/api/v1/authn'

        response = requests.post(url, data=json.dumps(payload), headers=headers)
        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            resp = json.loads(response.text)
            return resp
        else:
            response.raise_for_status()

    def __saml_assertion_aws(self, session_token, configuration):
        ''' fetch saml 2.0 assertion '''
        response = self.__okta_app_response(
            session_token=session_token,
            configuration=configuration
        )

        if self.verbose:
            Common.dump_verbose(
                message='SAML response: {}'.format(response.content)
            )

        soup = BeautifulSoup(response.content, "html.parser")
        assertion = None
        for inputtag in soup.find_all('input'):
            if inputtag.get('name') == 'SAMLResponse':
                assertion = inputtag.get('value')
        if not assertion:
            if self.verbose:
                Common.dump_verbose('Expecting \'<input name="SAMLResponse" value="...">\' in Okta response, but not found.')
            Common.dump_err(
                message='Unexpected response from Okta.',
                exit_code=4,
                verbose=self.verbose
            )
        return assertion

    def __okta_app_response(self, session_token, configuration):
        url = configuration['okta_aws_app_url'] + '?onetimetoken=' + session_token
        response = requests.get(url)
        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            return response
        response.raise_for_status()



''' ConfigGenerator class must be instantiated prior to use '''
import copy
import getpass
import json
import os

import click

from clokta.common import Common


class ConfigGenerator(object):
    ''' Loads or otherwise generates configuration information '''

    config_fields = [
        {
            'name': 'okta_username',
            'required': True,
            'save_to': 'default'
        },{
            'name': 'okta_password',
            'required': True,
            'secret': True
        },{
            'name': 'okta_org',
            'required': True,
            'save_to': 'default',
            'default_value': 'washpost.okta.com'
        },{
            'name': 'multifactor_preference',
            'required': True,
            'save_to': 'default',
            'default_value': "Okta Verify with Push",
            'prompt': 'Enter a preferred MFA - choose between Google Authenticator, SMS text message, Okta Verify, or Okta Verify with Push'
        },{
            'name': 'okta_aws_app_url',
            'required': True,
            'save_to': 'profile'
        },{
            'name': 'okta_password',
            'required': True,
            'secret': True
        },{
            'name': 'okta_aws_role_to_assume'
        },{
            'name': 'okta_onetimepassword_secret',
            'secret': True
        }
    ]

    @classmethod
    def generate_configuration(cls, config_section, verbose=False):
        configuration = {}
        for field in ConfigGenerator.config_fields:
            key = field['name']
            is_secret = 'secret' in field and field['secret']
            from_env = os.getenv(key=key, default=-1)
            if from_env != -1:
                # If defined in environment, use that first
                configuration[key] = from_env
            elif key in config_section and config_section[key]:
                # If defined in the config file, make sure it's not a secret, otherwise use it
                if is_secret:
                    Common.dump_err(
                        message='Invalid configuration.  {} should never be defined in clokta.cfg.'.format(key),
                        exit_code=6,
                        verbose=False)
                configuration[key] = config_section[key]
            elif 'required' in field and field['required']:
                    # We need it.  Prompt for it.
                    prompt = field['prompt'] if 'prompt' in field and field['prompt'] is not '' else 'Enter a value for {}'.format(key)
                    if 'secret' in field and field['secret']:
                        configuration[key] = getpass.getpass(prompt=prompt+":")
                    else:
                        configuration[key] = click.prompt(prompt, type=str, default=field['default_value'] if 'default_value' in field else None)

        if verbose:
            copy_config = copy.deepcopy(configuration)
            for key in copy_config:
                if key.endswith('password'):
                    copy_config[key] = '<redacted>'
            msg = 'Configuration: {}'.format(
                json.dumps(obj=copy_config, indent=4)
            )
            Common.dump_verbose(message=msg)

        return configuration



''' ProfileManager class must be instantiated prior to use. '''
import click
import configparser
import enum
import json
import os

from clokta.common import Common
from clokta.config_generator import ConfigGenerator


class OutputFormat(enum.Enum):
    ''' Enumeration for credentials output format codes '''
    ShellScript = 'SHELL'
    Profile = 'PROFILE'
    All = 'ALL'


class ProfileManager(object):
    ''' Supports profile file management '''

    def __init__(
        self,
        profile_name,
        config_location='~/.clokta/clokta.cfg',
        profiles_location='~/.aws/credentials',
        verbose=False
    ):
        ''' Instance constructor '''
        self.profile_name = profile_name
        self.verbose = verbose
        self.profiles_location = os.path.expanduser(profiles_location)
        self.config_location = os.path.expanduser(config_location)

    def initialize_configuration(self):
        ''' Generate and load config file section '''
        parser = configparser.ConfigParser()
        parser.read(self.config_location)

        if not parser['DEFAULT']:
            parser['DEFAULT'] = {
                'okta_org': ''
            }

        if self.profile_name not in parser.sections():
            msg = 'No profile "{}" in clokta.cfg, but enter the information and clokta will create a profile.\nCopy the link from the Okta App'
            app_url = click.prompt(msg.format(self.profile_name), type=str).strip()
            if not app_url.startswith("https://") or not app_url.endswith("?fromHome=true"):
                Common.dump_err("Invalid App URL.  URL usually of the form https://xxxxxxxx.okta.com/.../272?fromHome=true", 6, False)
            else:
                app_url = app_url[:-len("?fromHome=true")]
            parser[self.profile_name] = {
                'okta_aws_app_url': app_url
            }

        config_section = parser[self.profile_name]
        updated_config = ConfigGenerator.generate_configuration(
            config_section=config_section,
            verbose=self.verbose
        )
        self.__write_config(
            path_to_file=self.config_location,
            parser=parser
        )
        return updated_config

    def update_configuration(self, profile_configuration):
        ''' Update a config file section '''
        parser = configparser.ConfigParser()
        parser.read(self.config_location)

        default_keys = [field['name'] for field in ConfigGenerator.config_fields if 'save_to' in field and field['save_to']=='default']
        for key in default_keys:
            parser['DEFAULT'][key] = profile_configuration.get(key)

        profile_keys = [field['name'] for field in ConfigGenerator.config_fields if 'save_to' in field and field['save_to']=='profile']
        for key in profile_keys:
            parser[self.profile_name][key] = profile_configuration[key]

        if self.verbose:
            Common.dump_verbose(
                message='Re-writing configuration file {}'.format(self.config_location)
            )
        self.__write_config(
            path_to_file=self.config_location,
            parser=parser
        )

    def apply_credentials(self, credentials, echo_message=False):
        ''' Save a set of temporary credentials '''
        if self.verbose:
            msg = json.dumps(obj=credentials, default=Common.json_serial, indent=4)
            Common.dump_verbose(message=msg)

        parser = configparser.ConfigParser()
        parser.read(self.profiles_location)

        if self.profile_name not in parser.sections():
            if self.verbose:
                Common.dump_verbose(
                    message='Adding profile section {}'.format(self.profile_name)
                )
            parser.add_section(self.profile_name)

        creds = credentials['Credentials']
        parser[self.profile_name]['AWS_ACCESS_KEY_ID'] = creds['AccessKeyId']
        parser[self.profile_name]['AWS_SECRET_ACCESS_KEY'] = creds['SecretAccessKey']

        if 'AWS_SESSION_TOKEN' in parser[self.profile_name]:
            del parser[self.profile_name]['AWS_SESSION_TOKEN']

        if 'SessionToken' in creds:
            parser[self.profile_name]['AWS_SESSION_TOKEN'] = creds['SessionToken']

        if self.verbose:
            Common.dump_verbose(
                message='Re-writing credentials file {}'.format(self.profiles_location)
            )

        self.__write_config(
            path_to_file=self.profiles_location,
            parser=parser
        )

    def __write_config(self, path_to_file, parser):
        ''' Write config to file '''
        self.__backup_file(path_to_file=path_to_file)

        if not os.path.exists(os.path.dirname(path_to_file)):
            os.makedirs(os.path.dirname(path_to_file))

        with open(path_to_file, 'w') as file:
            parser.write(file)

    def __backup_file(self, path_to_file):
        ''' Back up config '''
        backup_location = os.path.expanduser(
            '{}.bak'.format(
                path_to_file
            )
        )
        if os.path.isfile(path=path_to_file):
            with open(path_to_file, 'r') as file:
                contents = file.read()

            with open(backup_location, 'w') as bak_file:
                bak_file.write(contents)

    def write_sourceable_file(self, credentials):
        '''
        Generates a shell script to source in order to apply credentials to the shell environment.
        '''
        creds = credentials['Credentials']
        output_file_name = '{dir}/{profile}.sh'.format(
            dir=os.path.dirname(self.config_location),
            profile=self.profile_name
        )
        lines = [
            'export AWS_ACCESS_KEY_ID={}\n'.format(creds['AccessKeyId']),
            'export AWS_SECRET_ACCESS_KEY={}\n'.format(creds['SecretAccessKey'])
        ]
        if 'SessionToken' in creds:
            lines.append('export AWS_SESSION_TOKEN={}\n'.format(creds['SessionToken']))
        else:
            lines.append('unset AWS_SESSION_TOKEN')

        with open(output_file_name, mode='w') as file_handle:
            file_handle.writelines(lines)

        return output_file_name

    def write_dockerenv_file(self, credentials):
        '''
        Generates a Docker .env file that can be used with docker compose to inject into the environment.
        '''
        creds = credentials['Credentials']
        output_file_name = '{dir}/{profile}.env'.format(
            dir=os.path.dirname(self.config_location),
            profile=self.profile_name
        )
        lines = [
            'AWS_ACCESS_KEY_ID={}\n'.format(creds['AccessKeyId']),
            'AWS_SECRET_ACCESS_KEY={}\n'.format(creds['SecretAccessKey'])
        ]
        if 'SessionToken' in creds:
            lines.append('AWS_SESSION_TOKEN={}\n'.format(creds['SessionToken']))

        with open(output_file_name, mode='w') as file_handle:
            file_handle.writelines(lines)

        return output_file_name




''' RoleChooser class must be instantiated prior to use '''
import base64
import xml.etree.ElementTree as ET

import click

from clokta.common import Common
from clokta.factors import Factors


class RoleChooser(object):
    ''' Supports AWS Role determination '''

    def __init__(self, saml_assertion, role_preference=None, verbose=False):
        ''' Instance constructor '''
        self.saml_assertion = saml_assertion
        self.role_preference = role_preference
        self.verbose = verbose

    def choose_idp_role_tuple(self):
        ''' Determine the role options the user can choose from '''
        idp_role_tuples = self.__discover_role_idp_tuples()

        # throw an error if no roles are provided
        #  (defensive coding only - this should not be impossible)
        if not idp_role_tuples:
            Common.dump_err(
                message='No AWS Role was assigned to this application!',
                exit_code=4,
                verbose=self.verbose
            )

        # use the one prvided if there is only one
        if len(idp_role_tuples) == 1:
            role_arn = idp_role_tuples[0][2]
            if self.role_preference and role_arn != self.role_preference:
                Common.echo(
                    message='Your cofigured role was not found; using {role}'.format(
                        role=role_arn
                    )
                )
            else:
                Common.echo(
                    message='Using the configured role {role}'.format(
                        role=role_arn
                    )
                )
            return idp_role_tuples[0]

        # use the configured role if it matches one from the the SAML assertion
        for tup in idp_role_tuples:
            if tup[2] == self.role_preference:
                return tup

        # make the user choose
        return self.__choose_tuple(idp_role_tuples=idp_role_tuples)

    def __discover_role_idp_tuples(self):
        ''' Generate tuples from each possible IDP/Role combination in the SAML assertion '''
        idp_role_pairs = []
        decoded_assertion = base64.b64decode(self.saml_assertion)
        root = ET.fromstring(decoded_assertion)
        for saml2_attribute in root.iter('{urn:oasis:names:tc:SAML:2.0:assertion}Attribute'):
            if saml2_attribute.get('Name') == 'https://aws.amazon.com/SAML/Attributes/Role':
                for saml2_attribute_value in saml2_attribute.iter('{urn:oasis:names:tc:SAML:2.0:assertion}AttributeValue'):
                    idp_role_pairs.append(saml2_attribute_value.text)

        idp_role_tuples = []
        for index, pair in enumerate(idp_role_pairs):
            idp_role = pair.split(',')
            tup = (index + 1), idp_role[0], idp_role[1]
            idp_role_tuples.append(tup)

        return idp_role_tuples

    def __choose_tuple(self, idp_role_tuples):
        ''' Give the user a choice from the intersection of configured and supported factors '''
        index = 1
        for tup in idp_role_tuples:
            slashIndex = tup[2].find('/')
            shortName = tup[2][slashIndex+1:] if slashIndex >= 0 else tup[2]
            msg = '{index} - {prompt}'.format(
                index=tup[0],
                prompt=shortName
            )
            Common.echo(message=msg, bold=True)
            index += 1

        raw_choice = None
        try:
            raw_choice = click.prompt('Choose a Role ARN to use', type=int)
            choice = raw_choice - 1
        except ValueError:
            Common.echo(message='Please select a valid option: you chose: {}'.format(raw_choice))
            return self.__choose_tuple(idp_role_tuples=idp_role_tuples)

        if len(idp_role_tuples) > choice >= 0:
            pass
        else:
            Common.echo(message='Please select a valid option: you chose: {}'.format(raw_choice))
            return self.__choose_tuple(idp_role_tuples=idp_role_tuples)

        chosen_option = idp_role_tuples[choice]
        if self.verbose:
            Common.dump_verbose(
                message='Using chosen Role {role} & IDP {idp}'.format(
                    role=tup[2],
                    idp=tup[1]
                )
            )

        return chosen_option



'''
Simple utility methods for the module
'''
import logging
import sys
from datetime import date, datetime

import click


class Common(object):
    ''' Console printing helpers '''

    @classmethod
    def dump_err(cls, message, exit_code, verbose=False):
        ''' Dump error to console and exit program '''
        if verbose:
            logging.exception(message)
        click.secho(message, bold=True, fg='red')
        sys.exit(exit_code)

    @classmethod
    def dump_verbose(cls, message, new_line=True, bold=False):
        ''' Dump verbose message to console '''
        click.secho(message, nl=new_line, bold=bold, fg='blue')

    @classmethod
    def echo(cls, message, new_line=True, bold=False):
        ''' Dump standard message to console '''
        click.secho(message, nl=new_line, bold=bold)

    @classmethod
    def json_serial(cls, obj):
        ''' JSON serializer for objects not serializable by default json code '''
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        raise TypeError("Type %s not serializable" % type(obj))



'''
This is the entry-point to the cli application.
'''
import click

from clokta.role_assumer import RoleAssumer


@click.command()
@click.option('--verbose', '-v', is_flag=True, help='Show detailed')
@click.option('--profile', '-p', required=True, help='Configuration profile')
def assume_role(profile, verbose=False):
    ''' Click point of entry '''
    assumer = RoleAssumer(profile=profile, verbose=verbose)
    assumer.assume_role()
